import pandas as pd
import os
from pkg_resources import resource_filename
from sklearn.model_selection import train_test_split

def lending_club(file_name = 'sample_credit_data.pkl', modelling_mode = True):
    """Sample test loan default data from Lending Club https://www.lendingclub.com/. Only a sample of all loans and available features is provided

    Metadata:
        id                         object - Loan ID
        loan_issue_date    datetime64[ns] - Date when loan was issued. All dates are set to fist day of the month
        default                     int64 - Flag if the loan went to default or not. 
                                            Default is defined as one of statuses: 'Late (31-120 days)', 'Late (16-30
                                            days)', 'Default'
        loan_amnt                 float64 - Amount requested
        funded_amnt               float64 - Amount granted
        term                        int64 - Term of the loan in months
        int_rate                  float64 - Annual interest rate
        annual_inc                float64 - Annual income of the client
        fico_range_low            float64 - Fico rate range low
        fico_range_high           float64 - Fico rate range high
        long_emp                    int64 - Indicator if client is employed for longer than 5 years
        credit_grade_A              uint8 - Indicator if clients credit grade is A
        credit_grade_B              uint8 - Indicator if clients credit grade is B
        credit_grade_C              uint8 - Indicator if clients credit grade is C
        credit_grade_D              uint8 - Indicator if clients credit grade is D
        credit_grade_E              uint8 - Indicator if clients credit grade is E
        credit_grade_F              uint8 - Indicator if clients credit grade is F
        credit_grade_G              uint8 - Indicator if clients credit grade is G

    Args:
        file_name (str) : name of the file which will be loaded from the data folder
        modelling_mode (bool) : True if you to get x_train, x_test, y_train, y_test

    Returns:
        credit_df (pandas.DataFrame): DataFrame with loan defaults data from lending club
        X_train (array): features for the train set 
        X_test (array): features for the test set 
        y_train (array): targets for the train set 
        y_test (array): targets for the test set 

    """


    filepath = resource_filename("probatus", os.path.join('datasets/data',file_name))
    credit_df = pd.read_pickle(filepath)


    X_train = []
    X_test = []
    y_train = []
    y_test = []

    if modelling_mode:
        y = credit_df[['default']]
        X = credit_df.drop(['id', 'loan_issue_date','default'], axis = 1)
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.33, random_state=42, stratify = y)

    return credit_df, X_train, X_test, y_train, y_test



import pandas as pd
import numpy as np
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
import warnings

from .. utils import assure_numpy_array, UnsupportedModelError, class_name_from_object




class ResemblanceModel(object):

    def __init__(self, model_type='rf', keep_samples=False, **model_kwargs):
        """
        This model checks for similarity of two samples.

        It assignes to each sample a label, reshuffles the values and randomly selects a fraction of the sample mix for
        training the model, and the remainder to estimate how well it can distinguish the samples.

        If the samples are coming from the same distribution, you would expect an AUC on the test to be of the order
        of 0.5, i.e. it is not possible to distingusih the samples

        Args:
            model_type: (str or model) define which model you want to run for resemblance modeling.
            Currently supported:
                - 'rf' (default): RandomForestClassifier from sklearn
                - 'lr': LogisticRegression from sklearn

                You can define your custom model object:
                Example:
                    from xgboost import XGBoostClassifier

                    ResemblanceModel(model_type=XGBoostClassifier()).fit(X1,X2)

            keep_samples: boolean, default is False. Set it to True if you want to retrieve the samples used by
                the underlying model to fit. Those samples are build from merging the X1 and X2 samples and then randomly
                sampling.
                If true, the samples are available as properties:
                    - X_train: features used for training the underlying model
                    - y_train: label of the samples used
            **model_kwargs:
        """

        self.fitted = False
        self.model = None
        self._keep_samples = keep_samples
        self.col_names = None

        if model_type =='rf':
            self.model = RandomForestClassifier(**model_kwargs)
        elif model_type =='lr':
            self.model = LogisticRegression(**model_kwargs)
        else:
            self.model = model_type(**model_kwargs)


    def __repr__(self):
        repr_ = f"{self.__class__.__name__}\n\tUnderlying model type: {self.get_model_name()}"
        if self.fitted:
            repr_ += "\nThe model is able to distinguish the samples with an AUC of {:.3f}".format(self.auc_test)
        return repr_

    def get_model_name(self):
        return class_name_from_object(self.model)


    def _get_feature_importance(self):
        """Returns feature importance for a given model

        Returns:
            importances (pandas series): feature importance
        """

        try:
            importances = self.model.feature_importances_
        except:
            try:
                importances = self.model.coef_
                importances = importances.reshape(importances.shape[1],)

            except:
                raise UnsupportedModelError("Model type {} is not supported".format(self.get_model_name()))

        if self.col_names is not None:
            feat_importances = pd.Series(importances, index=self.col_names)
        else:
            feat_importances = pd.Series(importances)

        return feat_importances

    def _check_and_fill_missings(self, X, impute_method=True):
        """
        Checks if for missing values: if there are and impute_method=True
        it fills them with zero

        Args:
            X (numpy array):
            impute_method (bool)
        Returns:
            X (numpy array)
        """
        if np.isnan(X).any() == True:
            warnings.warn("You have missing values in your sample",
                          DeprecationWarning,
                          stacklevel=2)
        if impute_method:
            warnings.warn("Going to temporary solution: filling missing values with zero",
                          DeprecationWarning,
                          stacklevel=2)
            X[np.isnan(X)] = 0
        return X


    def _propensity_check(self, X1, X2, test_size=0.33, random_state=None):
        """Calculates and returns the AUC score.

        Args:
            X1 (numpy.ndarray): first sample
            X2 (numpy.ndarray): second sample
            test_size: size of the test sample (fraction)


        Returns:
            auc score on the training sample, auc on the test sample, and feature imporatances  : float
        """

        assert X1.shape[1] == X2.shape[1], "The number of columns in X1 and X2 does not match!"
        assert type(X1) == type(X2), "Non matching types for  X1 and X2"

        if type(X1)==pd.DataFrame:
            self.col_names = X1.columns
            assert X1.columns.tolist() == X2.columns.tolist(), "The columns of the DataFrame X1 and X2 do not match"

        X1 = assure_numpy_array(X1)
        X2 = assure_numpy_array(X2)

        X1 = self._check_and_fill_missings(X1)
        X2 = self._check_and_fill_missings(X2)

        X = np.concatenate([X1,X2])
        y = np.zeros(len(X1) + len(X2))
        y[len(X1):] = 1
        X_train, X_test, y_train, y_test = train_test_split(X,y,test_size=test_size,random_state=random_state)

        if self._keep_samples:
            self.X_train = X_train
            self.X_test = X_test
            self.y_train =  y_train
            self.y_test = y_test

        self.model.fit(X_train,y_train)


        auc_train = roc_auc_score(y_train, self.model.predict_proba(X_train)[:,1])
        auc_test = roc_auc_score(y_test, self.model.predict_proba(X_test)[:, 1])


        importances = self._get_feature_importance()

        return auc_train, auc_test, importances

    def fit(self, X1,X2, **kwargs):
        """
            Fit the underlying model to compare the samples X1 or X2


        Args:
            X1 (numpy.ndarray or pd.DataFrame): first sample
            X2 (numpy.ndarray or pd.DataFrame): second sample
            **kwargs:
                test_size (float, default=0.33): fraction to use for test samples
                random_state=None: random state of train test split

        """


        self.auc_train, self.auc_test, self.importances = self._propensity_check(X1,X2,**kwargs)
        self.fitted=True

        return self








import itertools
import warnings

import pandas as pd
from tqdm import tqdm

from probatus.binning import SimpleBucketer, AgglomerativeBucketer, QuantileBucketer
from probatus.stat_tests import es, ks, psi, ad, sw


class DistributionStatistics(object):
    """
    Wrapper that applies a statistical method and a binning strategy to data.

    Parameters
    ----------
    statistical_test: string
        Statistical method to apply, statistical methods implemented:
            'ES': Epps-Singleton
            'KS': Kolmogorov-Smirnov statistic
            'PSI': Population Stability Index
            'SW': Shapiro-Wilk based difference statistic
            'AD': Anderson-Darling TS


    binning_strategy: string or None
        Binning strategy to apply, binning strategies implemented:
            'SimpleBucketer': equally spaced bins
            'AgglomerativeBucketer': binning by applying the Scikit-learn implementation of Agglomerative Clustering
            'QuantileBucketer': bins with equal number of elements

    bin_count: integer or None
        In case binning_strategy is not None, specify the number of bins to be used by the binning strategy

    Example:
    d1 = np.histogram(np.random.normal(size=1000), 10)[0]
    d2 = np.histogram(np.random.normal(size=1000), 10)[0]

    myTest = DistributionStatistics('KS', 'SimpleBucketer', bin_count=10)
    myTest.fit(d1, d2, verbose=True)

    """
    statistical_test_list = ['ES', 'KS', 'PSI', 'AD', 'SW']
    binning_strategy_list = ['simplebucketer', 'agglomerativebucketer', 'quantilebucketer', None]

    def __init__(self, statistical_test, binning_strategy, bin_count=None):
        self.statistical_test = statistical_test.upper()
        self.binning_strategy = binning_strategy
        self.bin_count = bin_count
        self.fitted = False

        if self.statistical_test.upper() not in self.statistical_test_list:
            raise NotImplementedError(f"The statistical test should be one of {self.statistical_test_list}")
        elif self.statistical_test.upper() == 'ES':
            self._statistical_test_function = es
        elif self.statistical_test.upper() == 'KS':
            self._statistical_test_function = ks
        elif self.statistical_test.upper() == 'PSI':
            self._statistical_test_function = psi
        elif self.statistical_test.upper() == 'SW':
            self._statistical_test_function = sw
        elif self.statistical_test.upper() == 'AD':
            self._statistical_test_function = ad

        if self.binning_strategy:
            if self.binning_strategy.lower() not in self.binning_strategy_list:
                raise NotImplementedError(f"The binning strategy should be one of {self.binning_strategy_list}")
            if self.binning_strategy.lower() == 'simplebucketer':
                self.binner = SimpleBucketer(bin_count=self.bin_count)
            elif self.binning_strategy.lower() == 'agglomerativebucketer':
                self.binner = AgglomerativeBucketer(bin_count=self.bin_count)
            elif self.binning_strategy.lower() == 'quantilebucketer':
                self.binner = QuantileBucketer(bin_count=self.bin_count)

    def __repr__(self):
        repr_ = f"DistributionStatistics object\n\tstatistical_test: {self.statistical_test}"
        if self.binning_strategy:
            repr_ += f"\n\tbinning_strategy: {self.binning_strategy}\n\tbin_count: {self.bin_count}"
        else:
            repr_ += "\n\tNo binning applied"
        if self.fitted:
            repr_ += f"\nResults\n\tvalue {self.statistical_test}-statistic: {self.statistic}"
        if hasattr(self, 'p_value'):
            repr_ += f"\n\tp-value: {self.p_value}"
        return repr_

    def fit(self, d1, d2, verbose=False, **kwargs):
        """
        Fit the DistributionStatistics object to data; i.e. apply the statistical test

        Args:
            d1: distribution 1
            d2: distribution 2
            verbose:

        Returns: statistic value and p_value (if available, e.g. not for PSI)

        """
        if self.binning_strategy:
            self.binner.fit(d1)
            d1_preprocessed = self.binner.counts
            d2_preprocessed = self.binner.apply_bucketing(d2)
        else:
            d1_preprocessed, d2_preprocessed = d1, d2

        res = self._statistical_test_function(d1_preprocessed, d2_preprocessed, verbose=verbose, **kwargs)
        self.fitted = True
        if type(res) == tuple:
            self.statistic, self.p_value = res
            return self.statistic, self.p_value
        else:
            self.statistic = res
            return self.statistic


class AutoDist(object):
    """
    Class to automatically apply all implemented statistical distribution tests and binning strategies to (a
    selection of) features in two dataframes.

    Parameters
    ----------
    statistical_tests: string 'all' or list of strings with tests to apply
        Statistical tests to apply, statistical methods implemented:
            'ES': Epps-Singleton
            'KS': Kolmogorov-Smirnov statistic
            'PSI': Population Stability Index
            'AD': Anderson-Darling TS

    binning_strategies: string 'all' or list of strings with strategies to apply
        Binning strategy to apply, binning strategies implemented:
            'SimpleBucketer': equally spaced bins
            'AgglomerativeBucketer': binning by applying the Scikit-learn implementation of Agglomerative Clustering
            'QuantileBucketer': bins with equal number of elements
            None: no binning is applied

    bin_count: integer, None or list of integers
        bin_count value(s) to be used, note that None can only be used when no bucketing strategy is applied

    Example:
        df1 = pd.DataFrame(np.random.normal(size=(1000, 2)), columns=['feat_0', 'feat_1'])
        df2 = pd.DataFrame(np.random.normal(size=(1000, 2)), columns=['feat_0', 'feat_1'])

        myAutoDist = AutoDist(statistical_tests='all', binning_strategies='all', bin_count=[10, 20])
        res = myAutoDist.fit(df1, df2, columns=df1.columns)
    """

    def __init__(self, statistical_tests='all', binning_strategies='all', bin_count=10):
        self.fitted = False
        if statistical_tests == 'all':
            self.statistical_tests = DistributionStatistics.statistical_test_list
        elif isinstance(statistical_tests, str):
            self.statistical_tests = [statistical_tests]
        else:
            self.statistical_tests = statistical_tests
        if binning_strategies == 'all':
            self.binning_strategies = DistributionStatistics.binning_strategy_list
        elif isinstance(binning_strategies, str):
            self.binning_strategies = [binning_strategies]
        else:
            self.binning_strategies = binning_strategies
        if not isinstance(bin_count, list):
            self.bin_count = [bin_count]
        else:
            self.bin_count = bin_count

    def __repr__(self):
        repr_ = "AutoDist object"
        if not self.fitted:
            repr_ += f"\n\tAutoDist not fitted"
        if self.fitted:
            repr_ += f"\n\tAutoDist fitted"
        repr_ += f"\n\tstatistical_tests: {self.statistical_tests}"
        repr_ += f"\n\tbinning_strategies: {self.binning_strategies}"
        repr_ += f"\n\tbin_count: {self.bin_count}"
        return repr_

    def fit(self, df1, df2, column_selection, return_failed_tests=True, suppress_warnings=True):
        """
        Fit the AutoDist object to data; i.e. apply the statistical tests and binning strategies

        Args:
            df1: dataframe 1 for distribution comparison with dataframe 2
            df2: dataframe 2 for distribution comparison with dataframe 1
            column_selection: list of columns in df1 and df2 that should be compared
            return_failed_tests: remove tests in result that did not succeed
            suppress_warnings: whether to suppress warnings during the fit process

        Returns: dataframe with results of the performed statistical tests and binning strategies

        """
        # test if all columns in column_selection are in df1 and df2
        if len(set(column_selection) - set(df1.columns)) or len(set(column_selection) - set(df2.columns)):
            raise Exception('Not all columns in `column_selection` are in the provided dataframes')

        result_all = pd.DataFrame()
        for col, stat_test, bin_strat, bins in tqdm(
                list(itertools.product(column_selection, self.statistical_tests, self.binning_strategies, self.bin_count))):
            dist = DistributionStatistics(statistical_test=stat_test, binning_strategy=bin_strat, bin_count=bins)
            try:
                if suppress_warnings:
                    warnings.filterwarnings('ignore')
                _ = dist.fit(df1[col], df2[col])
                if suppress_warnings:
                    warnings.filterwarnings('default')
                statistic = dist.statistic
                if hasattr(dist, 'p_value'):
                    p_value = dist.p_value
                else:
                    p_value = None
            except:
                statistic, p_value = 'an error occurred', None
                pass
            result_ = {'column': col, 'statistical_test': stat_test, 'binning_strategy': bin_strat, 'bin_count': bins,
                       'statistic': statistic, 'p_value': p_value}
            result_all = result_all.append(result_, ignore_index=True)
        if not return_failed_tests:
            result_all = result_all[result_all['statistic'] != 'an error occurred']
        self.fitted = True
        self._result = result_all[
            ['column', 'statistical_test', 'binning_strategy', 'bin_count', 'statistic', 'p_value']]
        self._result['bin_count'] = self._result['bin_count'].astype(int)
        self._result.loc[self._result['binning_strategy'].isnull(), 'binning_strategy'] = 'no_bucketing'

        # create pivot table as final output
        self.result = pd.pivot_table(self._result, values=['statistic', 'p_value'], index='column',
                                     columns=['statistical_test', 'binning_strategy', 'bin_count'], aggfunc='sum')
        # flatten multi-index
        self.result.columns = ["_".join([str(x) for x in line]) for line in self.result.columns.values]
        self.result.reset_index(inplace=True)
        return self.result



from .metric import get_metric
from .sampling import stratified_random
from .volatility import VolatilityEstimation


__all__ = ['get_metric', 'stratified_random', 'VolatilityEstimation', 'get_metric_folds']


import numpy as np
from sklearn.model_selection import train_test_split
from probatus.utils import assure_numpy_array

def get_metric(X, y, model, test_size, seed, evaluator, pred_type):
    """
    Draws random train/test sample from the data using random seed and calculates metric of interest.

    Args:
        X: nparray with features
        y: nparray with targets
        test_size: float fraction of data used for testing the model
        seed: int randomized seed used for splitting data
        evaluator : function used for calculating evaluation metric
        pred_type : string form of prediction which is used for obtaining evaluation metric.

    Returns: 
        list with metrics from tain, test and delta between the two

    """

    X = assure_numpy_array(X)
    y = assure_numpy_array(y)

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size = test_size, random_state = seed, stratify = y)
    model = model.fit(X_train, y_train)
    if pred_type == 'proba':
        metric_test = evaluator(y_test, model.predict_proba(X_test)[:,1])
        metric_train = evaluator(y_train, model.predict_proba(X_train)[:,1])
        metric_delta = metric_train - metric_test
    elif pred_type == 'class':
        metric_test = evaluator(y_test, model.predict(X_test))
        metric_train = evaluator(y_train, model.predict(X_train))
        metric_delta = metric_train - metric_test
    else:
        raise ValueError(f'Unsupported prediction type was passed: {pred_type}')

    return [metric_train, metric_test, metric_delta]

def get_metric_folds(x, y, model, x_test, y_test, evaluator, pred_type):
    """
    Draws random train/test independant samples from the data using random seed and calculates metric of interest.

    Args:
        X: nparray with features
        y: nparray with targets
        X_test: nparray test set with features
        y_test: nparray test set with targets        
        test_size: float fraction of data used for testing the model
        evaluator : function used for calculating evaluation metric
        pred_type : string form of prediction which is used for obtaining evaluation metric.

    Returns: 
        list with metrics from tain, test and delta between the two

    """    
    metrics_train = []
    metrics_test = []
    metrics_delta = []

    x = assure_numpy_array(x)
    y = assure_numpy_array(y)
    x_test = assure_numpy_array(x_test)
    y_test = assure_numpy_array(y_test)

    for train_x, train_y in zip(x, y):
        if np.mean(train_y) < 1 and np.mean(train_y) > 0:
            model = model.fit(train_x, train_y.flatten())

            if pred_type == 'proba':
                metric_test = evaluator(y_test, model.predict_proba(x_test)[:,1])
                metric_train = evaluator(train_y, model.predict_proba(train_x)[:,1])
                metric_delta = metric_train - metric_test
            elif pred_type == 'class':
                metric_test = evaluator(y_test, model.predict(x_test))
                metric_train = evaluator(train_y, model.predict(train_x))
                metric_delta = metric_train - metric_test
            else:
                raise ValueError(f'Unsupported prediction type was passed: {pred_type}')

            metrics_train.append(metric_train)
            metrics_test.append(metric_test)
            metrics_delta.append(metric_delta)

    return [np.mean(metrics_train),np.mean(metrics_test), np.mean(metrics_delta), np.var(metrics_train), np.var(metrics_test), np.var(metrics_delta)]


import pickle
import os
from pkg_resources import resource_filename

def lending_club_model(file_name = 'sample_credit_model.pkl'):
    """Sample Random Forest model trained on the lending club loan data.

    Model Hyper Parameberts:
        bootstrap=True, class_weight=None, criterion='gini',
        max_depth=6, max_features='auto', max_leaf_nodes=None,
        min_impurity_decrease=0.0, min_impurity_split=None,
        min_samples_leaf=1, min_samples_split=2,
        min_weight_fraction_leaf=0.0, n_estimators=100, n_jobs=None,
        oob_score=False, random_state=0, verbose=0, warm_start=False

    Args:
        file_name (str) : name of the file which will be loaded from the pretrained model

    Returns:
        credit_clf (sklearn.ensemple.RandomForestClassifier): Pretrained model

    """

    filepath = resource_filename("probatus", os.path.join('datasets/data',file_name))
    credit_clf = pickle.load(open(filepath, 'rb'))

    return credit_clf


from sklearn.model_selection import train_test_split

def stratified_random(x, y, size):
    """
    Draws N random samples from the data to create new train/test splits and calculate metric of interest.
    After collecting multiple metrics per split, summary statistics of the metic are reported.
    This provides uncertainty levels around the metric if the train/test data is resampled.  

    Args:
        X: pdDataFrame or nparray with features
        y: pdDataFrame or nparray with targets
        size : int sample size

    Returns: 
        train and test splits

    """
    X_train, X_test, y_train, y_test = train_test_split( x, y, test_size = size)

    return X_train, X_test, y_train, y_test


from scipy import stats
from ..utils import assure_numpy_array


def ad(d1, d2, verbose=False):
    """
    Calculates the Anderson-Darling TS on 2 distributions.

    Can be used on continuous or discrete distributions. Any binning/bucketing of the distributions/samples should be
    done before passing them to this function.

    Anderson & Darling 1954

    Advantages:
    - Unlike the KS, the AD (like the ES) can be used on both continuous & discrete distributions.
    - Works well even when dist has fewer than 25 observations.
    - More powerful than KS, especially for differences in the tails of distributions.

    Args:
        d1 (np.array or pandas.core.series.Series) : first sample
        d2 (np.array or pandas.core.series.Series) : second sample
        verbose (bool) : helpful interpretation msgs printed to stdout (default False)

    Returns:
        ad (float)     : AD test stat
        pvalue (float) : P value of rejecting the null hypothesis (that the two distributions are identical)
    """

    d1 = assure_numpy_array(d1)
    d2 = assure_numpy_array(d2)

    ad, critical_values, pvalue = stats.anderson_ksamp([d1, d2])

    if verbose:
        print('\nAD: pvalue =', pvalue)

        if pvalue < 0.01:
            print('\nAD: Null hypothesis rejected with 99% confidence. Distributions very different.')
        elif pvalue < 0.05:
            print('\nAD: Null hypothesis rejected with 95% confidence. Distributions different.')
        else:
            print('\nAD: Null hypothesis cannot be rejected. Distributions not statistically different.')

    return ad, pvalue



import numpy as np
from sklearn.model_selection import StratifiedKFold

def max_folds(y):
    """
    Calculate maximum number of stratified folds whih can be created from the data

    Args:
        y: nparray with targets

    Returns: 
        k: int max number of stratitfied folds

    """
    k = int(np.floor(y.shape[0]/(y.shape[0] * np.mean(y))))

    return k

def slicer(x, y, k):
    """
    Slice data into k independent folds

    Args:
        X: pdDataFrame or nparray with features
        y: pdDataFrame or nparray with targets
        k : int number of folds
    Returns: 
        x_slices: nparray sliced independent features folds
        y_slices: nparray sliced independent targets folds

    """    
    skf = StratifiedKFold(n_splits=k, random_state=None, shuffle=False)
    
    folds_train = []
    folds_test = []

    for train_index, test_index in skf.split(x, y):
        folds_train.append(train_index)
        folds_test.append(test_index) 
        
    indexes = np.append(folds_train[0],folds_test[0])
    x = x[indexes]
    y = y[indexes]
    
    x_slices = np.array_split(x,k)
    y_slices = np.array_split(y,k)
    return x_slices, y_slices


from .exceptions import NotFittedError, DimensionalityError, UnsupportedModelError
from .arrayfuncs import assure_numpy_array, check_1d
from .warnings import ApproximationWarning
from ._utils import class_name_from_object

__all__ = ['NotFittedError', 'DimensionalityError', 'UnsupportedModelError', 'assure_numpy_array', 'check_1d',
           'ApproximationWarning','class_name_from_object']



import numpy as np
import matplotlib.pyplot as plt
from sklearn.calibration import CalibratedClassifierCV
from sklearn.linear_model import LogisticRegression
from probatus.utils import assure_numpy_array
from probatus.binning import AgglomerativeBucketer, SimpleBucketer, QuantileBucketer

class Calibrator(object):

    def __init__(self, methods, folds, bucket_method, bins):
        """
        Initiate the calibrator

        Args:
            methods (dict) : list of calibration methods which should be applied, keys are names of the method 
                             to be applied. The values assigned to keys are the powers used for fitting the model.
                             Power applied only for MLE callibration
            folds (int) : number of folds which should be used for training the calibrator
            bucket_method (str) : bucketing method used for producing calibration plots
            bins (int) : number of bins use in the calibration
            
        """           
        self.methods = methods
        self.folds = folds
        self.bins = bins
        self.cal_models = dict()
        if bucket_method == 'agglomerative':
            self.bucketer = AgglomerativeBucketer(bin_count = bins)
        elif bucket_method == 'simple':
            self.bucketer = SimpleBucketer(bin_count = bins)
        elif bucket_method == 'quantile':
            self.bucketer = QuantileBucketer(bin_count = bins)
        else:
            raise ValueError(f'Unknown method for binning the data {bucket_method}')

    def calibration_curve(self, y_test, y_test_predict_proba):
        """
        Calculate the perfecentage of true class and mean predicted proba per bin

        Args:
            y_test (np.array) : numpy array with target on test set
            y_test_predict_proba (np.array) : predicted probability on the test set

        Returns:
            fraction_of_positives (np.array) : fraction of positives in the bin
            mean_predicted_value (np.array) : mean predicted probabilitie in the bin
        """   

        self.bucketer.fit(y_test_predict_proba)
        buckets = np.digitize(y_test_predict_proba,self.bucketer.boundaries,right=True)

        fraction_of_positives = []
        mean_predicted_value = []

        for i in range(1,self.bins + 1):
            idnexes = np.argwhere(buckets == i)
            mean = np.mean(y_test_predict_proba[idnexes])
            frac = np.sum(y_test[idnexes])/idnexes.shape[0]
            fraction_of_positives.append(frac)
            mean_predicted_value.append(mean)

        return np.array(fraction_of_positives), np.array(mean_predicted_value)

    def mle_calibration_model(self, x_train, y_train, model, power):
        """
        Train a non linear calibrated MLE model using Logistic Regression

        Args:
            model (object) : a model object which has predict_proba method
            x_train (np.array) : numpy array with features on train set
            y_train (np.array) : numpy array with target on train set
            power (int) : power which should be used in the callibrator

        Returns:
            lr_cal
            
        """   
        predIT = model.predict_proba(x_train)[:,1]
        predIT = predIT.reshape(predIT.shape[0], 1)
        lr_cal = LogisticRegression()

        features_array = predIT
        for i in range(2, power + 1):
            features_array = np.concatenate((features_array, predIT**i),axis=1)

        lr_cal.fit(features_array, y_train)

        return lr_cal

    def train_calibrator(self, model, x_train, y_train, method, power):
        """
        Train a calibration model based on selected method(s)

        Args:
            method (str) : calibration method
            model (object) : a model object which has predict_proba method
            x_train (np.array) : numpy array with features on train set
            y_train (np.array) : numpy array with target on train set
            power (int) : power which should be used in the callibrator
        Returns:
            clf
            
        """
        if method in ['sigmoid','isotonic']:
            clf = CalibratedClassifierCV(model, cv = self.folds, method = method)
            clf.fit(x_train, y_train)
        elif method == 'nonliear':
            clf = self.mle_calibration_model(x_train, y_train, model, power)
        else:
            raise ValueError(f'Unknown method for colibration {method}')

        return clf


    def calibration_plot(self, x_test, y_test, method_name, model, calibration, power):
        """
        Populate the calibration plot canvas 

        Args:
            calibration (object) : calibrated model
            x_test (np.array) : numpy array with features on test set
            y_test (np.array) : numpy array with target on test set
            bins (int) : number of bins used for calculating the calibration metrics
            method_name (str) : calibration method
            model (object) : a model object which has predict_proba method
            power (int) : power which should be used in the callibrator
            
        """

        features_array = x_test

        if method_name == 'nonliear':
            x_test = model.predict_proba(x_test)[:,1]
            x_test = x_test.reshape(x_test.shape[0], 1)
            features_array = x_test
            for i in range(2, power + 1):
                features_array = np.concatenate((features_array,x_test**i),axis=1)            

        y_test_predict_proba = calibration.predict_proba(features_array)[:, 1]
        fraction_of_positives, mean_predicted_value = self.calibration_curve(y_test, y_test_predict_proba)
        plt.plot(mean_predicted_value, fraction_of_positives, 's-', label=f'Calibrated - {method_name} {power}')


    def fit(self, model, x_train, x_test, y_test, y_train):
        """
        Train calibrators which can be used for calibrating probabilities. Storing all the trained calibrators in the dictionary

        Args:
            model (object) : a model object which has predict_proba method
            x_train (np.array) : numpy array with features on train set
            y_train (np.array) : numpy array with target on train set
            x_test (np.array) : numpy array with features on test set
            y_test (np.array) : numpy array with target on test set            

        """
        self.model = model
        self.x_test = assure_numpy_array(x_test)
        self.y_test = assure_numpy_array(y_test)

        methods_list = list(self.methods.keys())

        for method_i in methods_list:
            power = self.methods[method_i]
            cal_clf = self.train_calibrator(model, assure_numpy_array(x_train), assure_numpy_array(y_train), method_i, power)
            self.cal_models[method_i] = cal_clf



    def plot(self):
        """
        Creting a calibration plot     
        """
        fig, ax = plt.subplots(1, figsize=(12, 6))

        y_test_predict_proba = self.model.predict_proba(self.x_test)[:, 1]
        fraction_of_positives, mean_predicted_value = self.calibration_curve(self.y_test, y_test_predict_proba)
        plt.plot(mean_predicted_value, fraction_of_positives, 's-', label=f'Calibrated - Original')

        methods_list = list(self.methods.keys())

        for method in methods_list:
            power = self.methods[method]
            cal_clf = self.cal_models[method]
            self.calibration_plot(self.x_test, self.y_test, method, self.model, cal_clf, power)

        
        plt.plot([0, 1], [0, 1], '--', color='gray')
        plt.gca().xaxis.set_ticks_position('none')
        plt.gca().yaxis.set_ticks_position('none')
        plt.gca().legend()
        plt.show()

    def score(self, method, x, model, power = None):
        """
        Calibrate the probabilities from the original model

        Args:
            method (string) : method used for calibration
            x (np.array) : model features which will be used for producing calibrated probabilities    
            model (object) : a model object which has predict_proba method
            power (int) : power which should be used in the callibrator
        Returns:
            cal_probas (np.array) : calibrated probabilities       
            
        """
    
        if method == 'nonliear':
            x = model.predict_proba(x)[:,1]
            x = x.reshape(x.shape[0], 1)
            for i in range(1, power + 1):
                x = np.concatenate((x,x**i),axis=1)             

        cal_clf = self.cal_models[method] 
        cal_probas = cal_clf.predict_proba(x)

        return cal_probas

    def get_calibs(self):
        """
        Returns calibrated models
        Returns:
            cal_models (dics) : dictionary with calibrated models           
        """

        return self.cal_models


import shap
import pandas as pd
import numpy as np
from ..utils import class_name_from_object, UnsupportedModelError


def shap_calc(model, X, approximate=False, **shap_kwargs):
    """
    Helper function to calculate the shapley values for a given model.
    Supported models for the moment are RandomForestClassifiers and XGBClassifiers
    In case the shapley values have
    Args:
        model: pretrained model (Random Forest of XGBoost at the moment)
        X (pd.DataFrame or np.ndarray): features set
        approximate: boolean, if True uses shap approximations - less accurate, but very fast
        **shap_kwargs: kwargs of the shap.TreeExplainer

    Returns: (np.ndarray) shapley_values for the model

    """

    model_name = class_name_from_object(model)

    supported_models = ['RandomForestClassifier', 'XGBClassifier']

    if model_name not in supported_models:
        raise UnsupportedModelError(
            "{} not supported. Please try one of the following {}".format(model_name, supported_models))

    explainer = shap.TreeExplainer(model, **shap_kwargs)

    # Calculate Shap values
    shap_values = explainer.shap_values(X, approximate=approximate)

    if isinstance(shap_values, list):
        shap_values = shap_values[1]

    return shap_values


def shap_to_df(model, X, **kwargs):
    """
    Calculates the shap values and return the pandas DataFrame with the columns and the index of the original

    Args:
        model: pretrained model (Random Forest of XGBoost at the moment)
        X (pd.DataFrame or np.ndarray): features set
        **kwargs: for the function shap_calc

    Returns:

    """

    shap_values = shap_calc(model, X, **kwargs)
    if isinstance(X,pd.DataFrame):
        return pd.DataFrame(shap_values, columns=X.columns, index=X.index)

    elif isinstance(X,np.ndarray) and len(X.shape)==2:
        return pd.DataFrame(shap_values, columns=["col_{}" for ix in range(X.shape[1])])

    else: raise NotImplementedError("X must be a dataframe or a 2d array")


def mean_shap(df):
    """
    Returns the average shapley value for each column of the dataframe, as well as the average absolute shap value

    Args:
        df: df of shapley values (output of function shap_to_df)

    Returns: tuple of pd.Series: average shap per column, average absolute value shap per column

    """

    average_shap = pd.Series(np.mean(df.values, axis=0), index=df.columns, name='average_shap')

    average_abs_shap = pd.Series(np.mean(np.abs(df.values), axis=0), index=df.columns, name='average_abs_shap')

    return average_shap, average_abs_shap


def mean_shap_raw(model, X, **kwargs):
    """
    Returns the average shapley value for each column of the dataframe, as well as the average absolute shap value.

    Args:
        model: pretrained model (Random Forest of XGBoost at the moment)
        X (pd.DataFrame or np.ndarray): features set
        **kwargs:
            approximate  (boolean) if True uses shap approximations - less accurate, but very fast
            **kwargs of the shap.TreeExplainer

    Returns: tuple of pd.Series: average shap per column, average absolute value shap per column

    """

    shap_df = shap_to_df(model, X, **kwargs)

    return mean_shap(shap_df)



import pandas as pd
import numpy as np
from sklearn.cluster import AgglomerativeClustering

from ..utils import NotFittedError
from ..utils import ApproximationWarning

import warnings


class Bucketer(object):
    def __init__(self):
        self.fitted = False

    def __repr__(self):
        repr_ = f"{self.__class__.__name__}\n\tbincount: {self.bin_count}"
        if self.fitted:
            repr_ += f"\nResults:\n\tcounts: {self.counts}\n\tboundaries: {self.boundaries}"
        return repr_

    def fit(self, x):
        """ Create bucketing on the array x

        Args:
            x: input array

        Returns: fitted bucketer object

        """
        self._fit(x)
        self.fitted = True
        return self

    def apply_bucketing(self, x_new):
        """
        Apply bucketing to new data

        Args:
            x_new: data to be bucketed

        Returns: counts of the elements in x_new using the bucketing that was obtained by fitting the Bucketer instance

        """
        if not self.fitted:
            raise NotFittedError('Bucketer is not fitted')
        else:
            # np.digitize returns the indices of the bins to which each value in input array belongs
            # the smallest value of the `boundaries` attribute equals the lowest value in the set the instance was
            # fitted on, to prevent the smallest value of x_new to be in his own bucket, we ignore the first boundary
            # value
            digitize_result = np.digitize(x_new, self.boundaries[1:], right=True)
            result = pd.DataFrame({'bucket': digitize_result}).groupby('bucket')['bucket'].count()
            # reindex the dataframe such that also empty buckets are included in the result
            result = result.reindex(np.arange(self.bin_count), fill_value=0)
            return result.values


class SimpleBucketer(Bucketer):
    """Create equally spaced bins using numpy.histogram function

    Usage:
    x = [1, 2, 1]
    bins = 3
    myBucketer = SimpleBucketer(bin_count=bins)
    myBucketer.fit(x)

    myBucketer.counts gives the number of elements per bucket
    myBucketer.boundaries gives the boundaries of the buckets

    """

    def __init__(self, bin_count):
        super().__init__()
        self.bin_count = bin_count

    @staticmethod
    def simple_bins(x, bin_count):
        counts, boundaries = np.histogram(x, bins=bin_count)
        return counts, boundaries

    def _fit(self, x):
        self.counts, self.boundaries = self.simple_bins(x, self.bin_count)


class AgglomerativeBucketer(Bucketer):
    """Create binning by applying the Scikit-learn implementation of Agglomerative Clustering

    Usage:
    x = [1, 2, 1]
    bins = 3
    myBucketer = AgglomerativeBucketer(bin_count=bins)
    myBucketer.fit(x)

    myBucketer.counts gives the number of elements per bucket
    myBucketer.boundaries gives the boundaries of the buckets

    """

    def __init__(self, bin_count):
        super().__init__()
        self.bin_count = bin_count

    @staticmethod
    def agglomerative_clustering_binning(x, bin_count):
        clustering = AgglomerativeClustering(n_clusters=bin_count).fit(np.asarray(x).reshape(-1, 1))
        df = pd.DataFrame({'x': x, 'label': clustering.labels_}).sort_values(by='x')
        cluster_minimum_values = df.groupby('label')['x'].min().sort_values().tolist()
        cluster_maximum_values = df.groupby('label')['x'].max().sort_values().tolist()
        # take the mean of the upper boundary of a cluster and the lower boundary of the next cluster
        boundaries = [np.mean([cluster_minimum_values[i + 1], cluster_maximum_values[i]]) for i in
                      range(len(cluster_minimum_values) - 1)]
        # add the lower boundary of the lowest cluster and the upper boundary of the highest cluster
        boundaries = [cluster_minimum_values[0]] + boundaries + [cluster_maximum_values[-1]]
        counts = df.groupby('label')['label'].count().values
        return counts, boundaries

    def _fit(self, x):
        self.counts, self.boundaries = self.agglomerative_clustering_binning(x, self.bin_count)


class QuantileBucketer(Bucketer):
    """Create bins with equal number of elements

    Usage:
    x = [1, 2, 1]
    bins = 3
    myBucketer = QuantileBucketer(bin_count=bins)
    myBucketer.fit(x)

    myBucketer.counts gives the number of elements per bucket
    myBucketer.boundaries gives the boundaries of the buckets

    """

    def __init__(self, bin_count):
        super().__init__()
        self.bin_count = bin_count

    @staticmethod
    def quantile_bins(x, bin_count, inf_edges=False):

        try:
            out, boundaries = pd.qcut(x, q=bin_count, retbins=True, duplicates='raise')
        except ValueError:
            # If there are too many duplicate values (assume a lot of filled missings)
            # this crashes - the exception drops them.
            # This means that it will return approximate quantile bins
            out, boundaries = pd.qcut(x, q=bin_count, retbins=True, duplicates='drop')
            warnings.warn(ApproximationWarning("Approximated quantiles - too many unique values" ))
        df = pd.DataFrame({'x': x})
        df['label'] = out
        counts = df.groupby('label').count().values.flatten()
        if inf_edges:
            boundaries[0] = -np.inf
            boundaries[-1] = np.inf
        return counts, boundaries

    def _fit(self, x):
        self.counts, self.boundaries = self.quantile_bins(x, self.bin_count)



from .ad import ad
from .es import es
from .ks import ks
from .psi import psi
from .sw import sw
from .distribution_statistics import DistributionStatistics, AutoDist

__all__ = ['ad', 'es', 'ks', 'psi', 'sw', 'DistributionStatistics', 'AutoDist']



from scipy import stats
from ..utils import assure_numpy_array


def ks(d1, d2, verbose=False):
    """
    Calculates the Kolmogorov-Smirnov statistic on 2 samples. Any binning/bucketing of the distributions/samples
    should be done before passing them to this function.

    Args:
        d1 (np.ndarray or pd.core.series.Series) : first sample
        d2 (np.ndarray or pd.core.series.Series) : second sample
        verbose (bool)                           : helpful interpretation msgs printed to stdout (default False)

    Returns:
        ks (float)     : KS test stat
        pvalue (float) : P value of rejecting the null hypothesis (that the two distributions are identical)
    """

    d1 = assure_numpy_array(d1)
    d2 = assure_numpy_array(d2)

    ks, pvalue = stats.ks_2samp(d1, d2)

    if verbose:
        print('\nKS: pvalue =', pvalue)

        if pvalue < 0.01:
            print('\nKS: Null hypothesis rejected with 99% confidence. Distributions very different.')
        elif pvalue < 0.05:
            print('\nKS: Null hypothesis rejected with 95% confidence. Distributions different.')
        else:
            print('\nKS: Null hypothesis cannot be rejected. Distributions not statistically different.')

    return ks, pvalue



import numpy as np
import matplotlib.pyplot as plt
from probatus.metric_uncertainty.metric import get_metric
from joblib import Parallel, delayed
from probatus.utils import assure_numpy_array
from probatus.metric_uncertainty import delong
from probatus.metric_uncertainty.utils import max_folds
from probatus.metric_uncertainty.sampling import stratified_random
from probatus.metric_uncertainty.utils import slicer
from probatus.metric_uncertainty.metric import get_metric_folds

class VolatilityEstimation(object):
    """
    Draws N random samples from the data to create new train/test splits and calculate metric of interest.
    After collecting multiple metrics per split, summary statistics of the metic are reported.
    This provides uncertainty levels around the metric if the train/test data is resampled.  

    Args:
        X: pdDataFrame or nparray with features
        y: pdDataFrame or nparray with targets
        evaluator : dict with name of the metric as a key and array holding [scoring function, scoring type]
            e.g. {'AUC' : [roc_auc_score,'proba'], 'ACC' : [accuracy_score,'class']}
        n_jobs: number of cores to be used. -1 sets it to use all the free cores
        method: str type of estimation which should be used, currently supports:
                boot_global - bootstrap replicates with global estimation of the AUC uncertainty with non onverlapping resampling
                boot_seed - boostrap replicates with local estimation of the AUC uncertainty (fixed split) and overlapping resampling
                delong - delong estimator of the AUC uncertainty

    """
    def __init__(self, model, X, y, evaluator, method, n_jobs=1):
        self.model = model
        self.X = assure_numpy_array(X)
        self.y = assure_numpy_array(y)
        self.evaluator = evaluator
        self.n_jobs = n_jobs
        self.metrics_list = {}
        self.method = method

    def estimate(self, test_prc, iterations = 1000):
        """
        Runs a parallel loop to sample multiple metrics using different train/test splits  

        Args:
            test_prc: flot percentage of data used for test partition
            iterations: int number of samples

        Returns: 
            Popultes dictionary with metrics data from sampling

        """
        metrics = list(self.evaluator.keys())

        for evaluator_i in metrics:

            if self.method == 'boot_seed':
                random_seeds = np.random.random_integers(0, 999999, iterations)
                results = Parallel(n_jobs=self.n_jobs)(delayed(get_metric)(self.X, 
                                                                           self.y, 
                                                                           self.model, 
                                                                           test_prc, 
                                                                           i, 
                                                                           self.evaluator[evaluator_i][0],
                                                                           self.evaluator[evaluator_i][1]) for i in random_seeds)
                self.metrics_list[evaluator_i] = np.array(results)
            
            elif self.method == 'boot_global':

                results = []
                X_train, X_test, y_train, y_test = stratified_random(self.X, self.y, test_prc)
                top_k = max_folds(y_train) 
                if top_k > 11:
                    top_k = 11
                for k in range(1,top_k + 1):

                    if k == 1:
                        x_slice = [X_train]
                        y_slice = [y_train]
                    
                    else:
                        x_slice, y_slice = slicer(X_train, y_train, k)

                    results_i = get_metric_folds(x_slice, 
                                                 y_slice, 
                                                 self.model, 
                                                 X_test, 
                                                 y_test,
                                                 self.evaluator[evaluator_i][0],
                                                 self.evaluator[evaluator_i][1])
                    results.append(results_i)
                self.metrics_list[evaluator_i] = np.array(results) 

            elif self.method == 'delong':    
                X_train, X_test, y_train, y_test = stratified_random(self.X, self.y, test_prc)
                model = self.model.fit(X_train, y_train)

                y_pred_test = model.predict_proba(X_test)[:,1]
                y_pred_train = model.predict_proba(X_train)[:,1]

                auc_train, auc_cov_train = delong.delong_roc_variance(y_train, y_pred_train)
                auc_test, auc_cov_test = delong.delong_roc_variance(y_test, y_pred_test)

                self.metrics_list[evaluator_i] = np.array([auc_train, auc_test,auc_train - auc_test, auc_cov_train, auc_cov_test, 0]).reshape(1,6)

    def reporting(self, metric):
        """
        Prints summary statistics of the metric 

        Args:
            metric: str name of the metric to report

        """        
        metric_data = self.metrics_list[metric]
        print(f'Mean of metric on train is {round(np.mean(metric_data[:,0]),2)}')
        print(f'Mean of metric on test is {round(np.mean(metric_data[:,1]),2)}')
        print(f'Mean of delta is {round(np.mean(metric_data[:,2]),2)}')

        if self.method == 'boot_seed':
            print(f'Standard Deviation of metric on train is {round(np.std(metric_data[:,0]),5)}')
            print(f'Standard Deviation of metric on test is {round(np.std(metric_data[:,1]),5)}')
            print(f'Standard Deviation of delta is {round(np.std(metric_data[:,2]),5)}')
        elif self.method == 'boot_global' or self.method == 'delong':
            print(f'Standard Deviation of metric on train is {round(np.mean(metric_data[:,3]),5)}')
            print(f'Standard Deviation of metric on test is {round(np.mean(metric_data[:,4]),5)}')           
            print(f'Standard Deviation of delta is {round(np.mean(metric_data[:,5]),5)}')

    def plot(self, metric):
        """
        Plots detribution of the metric 

        Args:
            metric: str name of the metric to report

        """  
        metric_data = self.metrics_list[metric]

        if self.method == 'boot_seed':
            train = metric_data[:,0]
            test = metric_data[:,1]
            delta = metric_data[:,2]
        elif self.method == 'boot_global' or self.method == 'delong':
            train = np.random.normal(np.mean(metric_data[:,0]), np.sqrt(np.mean(metric_data[:,3])), 10000)
            test = np.random.normal(np.mean(metric_data[:,1]), np.sqrt(np.mean(metric_data[:,4])), 10000)
            delta = np.random.normal(np.mean(metric_data[:,2]), np.sqrt(np.mean(metric_data[:,5])), 10000)

        plt.hist(train, alpha=0.5, label='Train')
        plt.hist(test, alpha=0.5, label='Test')
        plt.legend(loc='upper right')
        plt.show()

        plt.hist(delta, alpha=0.5, label='Delta')
        plt.legend(loc='upper right')
        plt.show()        


from scipy import stats
from ..utils import assure_numpy_array


def es(d1, d2, verbose=False):
    """
    Calculates the Epps-Singleton test statistic on 2 distributions. Can be used on continuous or discrete
    distributions. Any binning/bucketing of the distributions/samples should be done before passing them to this
    function.

    Whereas KS relies on the empirical distribution function, ES is based on the empirical characteristic function
    (Epps & Singleton 1986, Goerg & Kaiser 2009).

    Advantages:
    - Unlike the KS, the ES can be used on both continuous & discrete distributions.
    - ES has higher power (vs KS) in many examples.
    Disadvantages:
    - Not recommended for fewer than 25 observations. Instead, use the Anderson-Darling TS. (However, ES can still be
      used for small samples. A correction factor is applied so that the asymptotic TS distribution more closely follows
      the chi-squared distribution, such that p-values can be computed.)

    Args:
        d1 (np.array or pandas.core.series.Series) : first sample
        d2 (np.array or pandas.core.series.Series) : second sample
        verbose (bool) : helpful interpretation msgs printed to stdout (default False)

    Returns:
        es (float)     : ES test stat
        pvalue (float) : P value of rejecting the null hypothesis (that the two distributions are identical)
    """

    d1 = assure_numpy_array(d1)
    d2 = assure_numpy_array(d2)

    es, pvalue = stats.epps_singleton_2samp(d1, d2)

    if verbose:
        print('\nES: pvalue =', pvalue)

        if pvalue < 0.01:
            print('\nES: Null hypothesis rejected with 99% confidence. Distributions very different.')
        elif pvalue < 0.05:
            print('\nES: Null hypothesis rejected with 95% confidence. Distributions different.')
        else:
            print('\nES: Null hypothesis cannot be rejected. Distributions not statistically different.')

    return es, pvalue



import setuptools
import os

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

base_packages = ["scikit-learn>=0.20.2",
                 "pandas>=0.25",
                 "matplotlib==3.1.1",
                 "seaborn>=0.9.0",
                 "shap>=0.29",
		        "joblib>=0.13.2"]

try:
    if os.environ.get('CI_COMMIT_TAG'):
        version = os.environ['CI_COMMIT_TAG']
    else:
        version = os.environ['CI_JOB_ID']
except:
    version = 'local'


setuptools.setup(
    name='probatus',
    version=version,
    description='Tools for machine learning model validation',
    long_description=read('README.md'),
    author='RPAA ING',
    author_email='ml_risk_and_pricing_aa@ing.com',
    license='ING Open Source',
    packages=setuptools.find_packages(),
    package_data={'probatus': ['datasets/data/*.pkl']},
    install_requires=base_packages,
    url='https://gitlab.com/ing_rpaa/probatus',
    zip_safe=False
)



import numpy as np
import scipy.stats as stats
from ..utils import assure_numpy_array


def psi(d1, d2, verbose=False):
    """
    Calculates the Population Stability Index

    A simple statistical test that quantifies the similarity of two distributions. Commonly used in the banking / risk
    modeling industry. Only works on categorical data or bucketed numerical data. Distributions must be binned/bucketed
    before passing them to this function. Bin boundaries should be same for both distributions. Distributions must have
    same number of buckets. Note that the PSI varies with number of buckets chosen (typically 10-20 bins are used).

    Args:
        d1 (np.ndarray or pd.core.series.Series) : first distribution ("expected")
        d2 (np.ndarray or pd.core.series.Series) : second distribution ("actual")
        verbose (bool)                           : print useful interpretation info to stdout (default False)

    Returns:
        psi_value (float) : measure of the similarity between d1 & d2. (range 0-inf, with 0 indicating identical
                            distributions and > 0.25 indicating significantly different distributions)
    Raises:
        UserWarning: if number of bins in d1 or d2 is less than 10 or greater than 20, where PSI is not well-behaved.
    """

    d1 = assure_numpy_array(d1)
    d2 = assure_numpy_array(d2)

    if len(d1) < 10:
        raise UserWarning('PSI is not well-behaved when using less than 10 bins.')
    if len(d1) > 20:
        raise UserWarning('PSI is not well-behaved when using more than 10 bins.')

    if len(d1) != len(d2):
        raise ValueError('Distributions do not have the same number of bins.')

    # Number of bins/buckets
    b = len(d1)

    n = d1.sum()
    m = d2.sum()

    expected_ratio = d1 / n
    actual_ratio = d2 / m

    # Necessary to avoid divide by zero and ln(0). Should have minor impact on PSI value.
    for i in range(0, b):
        if expected_ratio[i] == 0:
            expected_ratio[i] = 0.0001
            if verbose:
                print(f"PSI: Bucket {i} has zero counts; may result in over-estimated (larger) PSI value. Decreasing \
                        the number of buckets may also help avoid buckets with zero counts.")
    for i in range(0, b):
        if actual_ratio[i] == 0:
            actual_ratio[i] = 0.0001
            if verbose:
                print(f"PSI: Bucket {i} has zero counts; may result in over-estimated (larger) PSI value. Decreasing \
                        the number of buckets may also help avoid buckets with zero counts.")

    psi_value = np.sum((actual_ratio - expected_ratio) * np.log(actual_ratio / expected_ratio))

    if verbose:
        print('\nPSI =', psi_value)

        print('\nPSI: Critical values defined according to de facto industry standard:')
        if psi_value <= 0.1:
            print('\nPSI <= 0.10: No significant distribution change.')
        elif 0.1 < psi_value <= 0.25:
            print('\nPSI <= 0.25: Small distribution change; may require investigation.')
        elif psi_value > 0.25:
            print('\nPSI > 0.25: Significant distribution change; investigate.')

        alpha = [0.95, 0.99, 0.999]
        z_alpha = stats.norm.ppf(alpha)
        psi_critvals = ((1 / n) + (1 / m)) * (b - 1) + z_alpha * ((1 / n) + (1 / m)) * np.sqrt(2 * (b - 1))
        print('\nPSI: Critical values defined according to Yurdakul 2018:')
        if psi_value > psi_critvals[2]:
            print('PSI: 99.9% confident distributions have changed.')
        elif psi_value > psi_critvals[1]:
            print('PSI: 99% confident distributions have changed.')
        elif psi_value > psi_critvals[0]:
            print('PSI: 95% confident distributions have changed.')
        elif psi_value < psi_critvals[0]:
            print('PSI: Distributions similar.')

    return psi_value



import numpy as np
import pandas as pd
import numbers

from probatus.utils import DimensionalityError


def check_1d(x):
    """
    Checks whether or not a list, numpy array, pandas dataframe, pandas series are one-dimensional.
    Returns True when check is ok, otherwise throws a `DimensionalityError`

    Args:
        x: list, numpy array, pandas dataframe, pandas series

    Returns: True or throws `DimensionalityError`

    """
    if isinstance(x, list):
        if any([isinstance(el, list) for el in x]):
            raise DimensionalityError('The input is not 1D')
        else:
            return True
    if isinstance(x, np.ndarray):
        if x.ndim == 1 and all([isinstance(el, numbers.Number) for el in x]):
            return True
        else:
            raise DimensionalityError('The input is not 1D')
    if isinstance(x, pd.core.frame.DataFrame):
        if len(x.columns) == 1 and pd.api.types.is_numeric_dtype(x[x.columns[0]]):
            return True
        else:
            raise DimensionalityError('The input is not 1D')
    if isinstance(x, pd.core.series.Series):
        if x.ndim == 1 and pd.api.types.is_numeric_dtype(x):
            return True
        else:
            raise DimensionalityError('The input is not 1D')


def assure_numpy_array(x, assure_1d=False):
    """
    Returns x as numpy array. X can be a list, numpy array, pandas dataframe, pandas series

    Args:
        x: list, numpy array, pandas dataframe, pandas series
        assure_1d: whether or not to assure that the input x is one-dimensional

    Returns: numpy array

    """
    if assure_1d:
        _ = check_1d(x)
    if isinstance(x, list):
        return np.array(x)
    if isinstance(x, np.ndarray):
        return x
    if isinstance(x, pd.core.frame.DataFrame):
        if len(x.columns) == 1:
            return x.values.flatten()
        else:
            return x.values
    if isinstance(x, pd.core.series.Series):
        return x.values
